# -*- coding: utf-8 -*-
"""
Created on Thu Jun  6 15:59:11 2019

@author: digvijaygusain

class for parameter sweep. Takes in World(), parameter names and values as lists.
Outputs nice figures based on seaborn for visualisation.
"""

import numpy as np, sys
import matplotlib.pyplot as plt
from time import sleep

class Sweep():
    '''
    With Sweep, you can perform parameter sweeps on cosimulation World created by FMUWorld.
    Sweep takes in three arguments:
        1. Cosimulation World object
        2. Sensitivity information
        3. Kind
        
    Sensitivity information is the most important parameter here since it defines what parameters should swept through and in what range.
    It is given as dictionary in the following form.
    
    1. Cosimulation World Object: World created using FMUWorld.
    
    2. sensitivity_info = {'parameters':[param1, param2, .., paramN],
                            'bounds':[[p1_l, p1_u, N*],
                                      [p2_l, p2_u, N*],
                                      [p3_l, p3_u, N*],
                                      [pN_l, pN_u, N*]],
                            'output_parameter': 'output_parameter_identifier'
                            }
        'parameters' lists the parameter names in the World object. These should be given as 'simulator_name.parameter_name'. simulator_name is the fmu, or powerflow object name in the World. 'parameter_name' is the parameter name, This can be further specified as object.variable. The definition follows a pattern similar to defining connections dictionary between World simulators.
        
        'bounds' specifies the bounds on each parameter listed. This is given as a numpy matrix of dimension Nx2, where N is the number of parameters specified. Each parameter has lower and upper bound specified as [p1_l, p1_u]. Additionally, Sweep accepts 'N' as a third value in this list which specifies the number of points between the bounds. This can be set differently for each parameter bounds. If not specified, a default value of 10 is chosen.  
        
        'output_parameter' specifies which parameter to look record. It is specified as per FMUWorld naming conventions. This should also be specified while creating the World object.
        
    3. Kind: specifies what type of plot should be created. Sweep supports contour plots for parametric (if two parameters are specified), and (default) simple line plots. All plots in defualt are given as multi axes line plots where x axes are parameters being varied, and y axis is the output_parameter. If kind = 'parametric', Sweep plots the output_parameter as a heat map, with x and y axes being the two parameters specified. For kind = 'parametric', the number of parameters must be exactly two.
        
    Once parameter sweep is completed, plots are generated and a result object is created. This can be exported as csv file with SweepObject.export_to_csv(csv_location).
    
    '''    
    
    
    def __init__(self, _cosimulation_world, _sensitivity_data, kind = 'single'):
        self.world = _cosimulation_world
        self.simulators_dict = {}
        self.simulators_dict.update(self.world.fmu_dict)
        self.simulators_dict.update(self.world.powerflow_dict)
        if len(self.world.powerflow_dict) > 0:
            self.powerflow_exists = True
        else:
            self.powerflow_exists = False
        self.info = _sensitivity_data
        self.kind = kind
        
        
    def sweep(self):
        if self.kind == 'single':
            res = self.single_parameter_sweep()
        elif self.kind == 'parametric':
            self.parametric_sweep()
        else:
            print('FMUWorld.Sweep currently supports only single and parametric sweeps.')
        return res
        
    def single_parameter_sweep(self):
        self.parameters_list = self.info['parameters']
#        print(f'param list = {self.parameters_list}')
        self.bounds_matrix = self.info['bounds']
        self.op_param = self.info['output_parameter']
        
        #process output parameter
        op_sim_name = self.op_param.split('.')[0]
        
        #re-order bounds matrix
        new_matrix = []
        for item in self.bounds_matrix:
            if len(item) == 2:
                new_item = np.linspace(item[0], item[1], 10)
            elif len(item) == 3:
                new_item = np.linspace(item[0], item[1], item[2])
            else:
                print('Bounds matrix should have elements with either two or three values to define lower, upper bounds and number of values.')
            new_matrix.append(new_item)
            
        new_matrix = np.array(new_matrix)        
        
        #store original parameters before changing anything
#        if self.powerflow_exists:
#            self.world.simulate()
        
        self.original_values = {}
        for parameter in self.parameters_list:
            sim_name = parameter.split('.')[0]
            sim_p = parameter.replace(sim_name,'')[1:]
            sim = self.simulators_dict[sim_name]
            tmp = sim.get_value([sim_p])
            self.original_values[parameter] = tmp[0]
        
        self.sweep_res = {}
        count = 0
        #create options list
        for parameter, value_list in zip(self.parameters_list, new_matrix):
            sim_name = parameter.split('.')[0]
            sim_param = parameter.replace(sim_name,'')[1:]
            self.sweep_res[sim_name] = {'data': [],
                                         'value': [],
                                         'y-axis': sim_name}
#            print(f"Simulation {round(count/len(self.parameters_list),1)}% complete.")
            cc = 0
            for value in list(value_list) + [self.original_values[parameter]]:
                
                self.my_options = {'init':{}} if len(self.world.init_dict) < 1 else {'init':self.world.init_dict}
                if sim_name not in self.my_options['init'].keys():
                    self.my_options['init'][sim_name] = ([], [])
                self.my_options['init'][sim_name][0].append(sim_param)
                self.my_options['init'][sim_name][1].append(value)
                self.world.options(self.my_options)
                results = self.world.simulate()
                
                a1 = results[op_sim_name].loc[:,['time', self.op_param]]
                self.sweep_res[sim_name]['data'].append(a1)
                self.sweep_res[sim_name]['value'].append(value)
                
                cc+=1
                if cc != len(list(value_list) + [self.original_values[parameter]]):
                    plt.figure(count)
                    plt.plot(a1['time'], a1[self.op_param], label=f'{parameter}: {value}')
                    plt.title(f"Sensitivity of {self.op_param} to {parameter}")
                    legend = plt.legend()
                    legend.draggable(state=True)
                            
            count+=1
        plt.show()
        
        return self.sweep_res
        
            
            
            
            
        
        
        
        
                
    def evaluation_function(self):
                
        
        
        
        
        pass        
    
    def parametric_sweep(self):
        pass
        
        



# -*- coding: utf-8 -*-
"""
Created on Sat Jul  6 15:07:17 2019

@author: Digvijay

Python script adapter for energysim
"""


import sys
class py_adapter():
    
    def __init__(self, script_name, script_loc, inputs = [], outputs = []):
        '''
        script_name should be something1.something2
        here: something1 is filename.py, and something2 is the function that is called from filename.
        something1.py should contain a python function something2 that can be imported as from something1 import something2.
        '''
        self.script_name = script_name
        self.script_loc = script_loc
        self.inputs = inputs
        self.outputs = outputs
        

    def init(self):
        sys.path.append(self.script_loc)
        a = self.script_name.split('.')[0]
        exec('import ' + a)
        
    
    def set_value(self, parameters, values):
        self.my_dict = dict(zip(parameters, values))

    
    def get_value(self, parameters):
        return self.outvalues
    
    def step(self):
        exec('import ' + self.script_name.split('.')[0])
#        exec('import ' + self.script_name.split('.')[0])
        self.outvalues = ''
#        print(f'executing: {self.outvalues} = {a}{self.script_name.split('.')[1]}({self.my_dict})')
#        self.outvalues = getattr(self.script_name.split('.')[0],f'{self.my_dict}')
        exec(f'self.outvalues ={self.script_name}({self.my_dict})')
#        pass
#    
    def getOutput(self):
        return self.outvalues
    
    
#    def set_value(self, parameters, values):
#        for parameter, value in zip(parameters, values):
#            ele_name, input_variable = parameter.split('.')
#            assert input_variable in ['P', 'Q'], "Powerflow input variable not valid. Use P, Q to  define variables."
#            if ele_name in list(self.network.gen.name):
#                adder, residual = 'gen', 'p_mw' if input_variable == 'P' else 'q_mvar'
#            elif ele_name in list(self.network.load.name):
#                adder, residual = 'load', 'p_mw' if input_variable == 'P' else 'q_mvar'
#            elif ele_name in list(self.network.sgen.name):
#                adder, residual = 'sgen', 'p_mw' if input_variable == 'P' else 'q_mvar'
#            else:
#                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
#                sys.exit()
#            
#            getattr(self.network, adder).at[(getattr(self.network, adder).name == ele_name).idxmax(), residual] = value
#    
#    def get_value(self, parameters):
#        temp_parameter_list = [x.split('.') for x in parameters]
#        temp_output = []
#        for ele_name, output_variable in temp_parameter_list:
#            if output_variable.lower() == 'v':
#                x = 'vm_pu'
#            elif output_variable.lower() == 'va':
#                x = 'va_degree'
#            if ele_name in list(self.network.gen.name):
#                adder, residual = 'res_gen', 'p_mw' if output_variable == 'P' else 'q_mvar'
#            elif ele_name in list(self.network.load.name):
#                adder, residual = 'res_load', 'p_mw' if output_variable == 'P' else 'q_mvar'
#            elif ele_name in list(self.network.bus.name):
#                adder, residual = 'res_bus', x
#            elif ele_name in list(self.network.sgen.name):
#                adder, residual = 'res_sgen', 'p_mw' if output_variable == 'P' else 'q_mvar'
#            else:
#                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
#                sys.exit()
#                
#            temp_var = getattr(self.network, adder).at[(getattr(self.network, adder[4:]).name == ele_name).idxmax(), residual]
#            temp_output.append(temp_var)
#        
#        return temp_output
#        
#    
#    def getOutput(self):
##        a1 = []
##        for item in self.new_outputs:
##            a1.append(self.get_value(item))
##        return a1    
#        return [getattr(self.network, adder).at[(getattr(self.network, adder[4:]).name == ele).idxmax(), residual] for ele, adder, residual in self.new_outputs]
#        
#    
#    def setInput(self, inputValues):
#        pass
    
    
        
    



# -*- coding: utf-8 -*-
"""
Created on Thu Jun  6 13:56:32 2019

@author: digvijaygusain

pypsa FMUWorld adapter
"""

import pypsa, sys, logging
pypsa.pf.logger.setLevel(logging.CRITICAL)

class pypsa_adapter():
    
    def __init__(self, network_name, net_loc, inputs = [], outputs = [], logger_level = 'CRITICAL'):
#        pypsa.pf.logger.setLevel(getattr(logging, logger_level))
        self.network_name = network_name
        self.net_loc = net_loc
        self.network = pypsa.Network()
        self.network.import_from_csv_folder(net_loc)
        print('Successfully imported network')
        assert len(self.network.snapshots) == 1, "Only one snapshot is supported."
        self.new_inputs, self.new_outputs = self.process_powerflow_ipop(self.network, inputs, outputs)
        self.outputs = outputs

    
    def init(self):
        self.network.lpf()
        self.network.pf(use_seed=True)
    
    def set_value(self, parameters, values):
        for parameter, value in zip(parameters, values):
            ele_name, input_variable = parameter.split('.')
            assert input_variable in ['P', 'Q'], "Powerflow input variable not valid. Use P, Q to  define variables."
            if ele_name in list(self.network.generators.index):
                adder, residual = 'generators', input_variable.lower()+'_set'
            elif ele_name in list(self.network.loads.index):
                adder, residual = 'loads', input_variable.lower()+'_set'
            elif ele_name in list(self.network.storage_units.index):
                adder, residual = 'storage_units', input_variable.lower()+'_set'
            else:
                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
                sys.exit()
            
            getattr(self.network, adder).at[ele_name, residual] = value
    
    def get_value(self, parameters):
        temp_parameter_list = [x.split('.') for x in parameters]
        temp_output = []
        for ele_name, output_variable in temp_parameter_list:
            if output_variable.lower() == 'v':
                x = 'v_mag_pu'
            elif output_variable.lower() == 'va':
                x = 'v_ang'
            if ele_name in list(self.network.generators.index):
                adder, residual = 'generators_t', output_variable.lower()
            elif ele_name in list(self.network.loads.index):
                adder, residual = 'loads_t', output_variable.lower()
            elif ele_name in list(self.network.buses.index):
                adder, residual = 'buses_t', x
            elif ele_name in list(self.network.storage_units.index):
                adder, residual = 'storage_units_t', output_variable.lower()
            else:
                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
                sys.exit()
                
            temp_var = getattr(getattr(self.network, adder), residual).at[list(getattr(getattr(self.network, adder), residual).index)[0], str(ele_name)]
            temp_output.append(temp_var)
        
        return temp_output
        
    
    def getOutput(self):
        return [getattr(getattr(self.network, adder), residual).at[list(getattr(getattr(self.network, adder), residual).index)[0], str(ele)] for ele, adder, residual in self.new_outputs]
        
    
    def setInput(self, inputValues):
        pass
    
    def step(self):  
#        pypsa.pf.logger.setLevel(logging.CRITICAL)
        self.network.lpf()
        self.network.pf(use_seed=True)
    
    def process_powerflow_ipop(self, network, inputs, outputs):
            new_inputs = []
            new_outputs = []
            for item in inputs:
                ele_name, input_variable = item.split('.')
                assert input_variable in ['P', 'Q'], "Powerflow input variable not valid. Use P, Q to  define variables."
                #check in generators
                if ele_name in list(network.generators.index):
                    adder, residual = 'generators', input_variable.lower()+'_set'
                elif ele_name in list(network.loads.index):
                    adder, residual = 'loads', input_variable.lower()+'_set'
                elif ele_name in list(network.storage_units.index):
                    adder, residual = 'storage_units', input_variable.lower()+'_set'
                else:
                    print(f'Only Generator, load, and storage P, Q inputs are supported. Couldnt find {ele_name} in either loads or generators. Quitting simulation.')
                    sys.exit()
                new_inputs.append((ele_name, adder, residual))
#                print('done input analysing')
            
            for item in outputs:
                ele_name, output_variable = item.split('.')
                assert output_variable in ['P', 'Q', 'V', 'Va'], "Powerflow output variable not valid. Use P, Q, V, Va to  define variables."
                
                if output_variable == 'V':
                    x = 'v_mag_pu'
                elif output_variable == 'Va':
                    x = 'v_ang'
                    
                if ele_name in list(network.generators.index):
                    adder, residual = 'generators_t', output_variable.lower()
                elif ele_name in list(network.loads.index):
                    adder, residual = 'loads_t', output_variable.lower()
                elif ele_name in list(network.buses.index):
                    adder, residual = 'buses_t', x
                elif ele_name in list(network.storage_units.index):
                    adder, residual = 'storage_units_t', input_variable.lower()
                else:
                    print(f'Only Generator, load, storage, and bus P, Q, outputs are supported. Couldnt find {ele_name} in specified network. Quitting simulation.')
                    sys.exit()
                new_outputs.append((ele_name, adder, residual))
#                print('done output analysing')
                
            return new_inputs, new_outputs   
    
    def cleanUp(self):
        del self.network


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 10 19:08:54 2018

@author: digvijaygusain
"""

__version__ = '1.1.1'
__author__ = 'Digvijay Gusain'

from .csAdapter import FmuCsAdapter
from .meAdapter import FmuMeAdapter
from .pypsaAdapter import pypsa_adapter
from .signalAdapter import signal_adapter
from .ppAdapter import pp_adapter
from .pyScriptAdapter import py_adapter
from fmpy.model_description import read_model_description
import sys
import numpy as np
import pandas as pd
import pypsa, logging as lg
from time import time as current_time, sleep
import networkx as nx
import matplotlib.pyplot as plt
from tqdm import tqdm
pypsa.pf.logger.setLevel(lg.CRITICAL)


'''
    World is your cosimulation canvas. You initialise World with basic simulation 
    parameters such as:
        1. start_time (=0 by default) 
        2. stop_time (=100 by default)
        3. logging (False by default). Gives users update on simulation progress
        4. exchange (=0 by dfault). Users can specify instance of information exchange
        5. clean_up (True by default) cleans up temporary files created during the simulation (can be very large in some cases, hence recommended to keep it True)
        6. interpolate_results (True by default) FMUs with larger time steps have their outputs interpolated to match the time step of the lowest time step of specified FMUs.
    
    After specifying the world, you can add FMUs to the world by add_fmu(). This is done by specifying the following:
        1. Valid fmu_name. It is extremely essential that you assign it relevant name. This helps to specify connections later.
        2. Valid fmu_location. Use a raw string address of FMU file
        3. outputs: specify the outputs that need to be observed. 
        4. step_size (=1e-3 by default)
    
    If required, signals can be added as well. This can be done via the add_signal command.
    <explain add_signal>
    The connections can be specified with a dictionary. It can be specified as:
        {'_fmu1.output_variable1':'_fmu2.input_variable1',
        '_fmu1.output_variable2':'_fmu2.input_variable2',
        '_fmu1.output_variable3':'_fmu2.input_variable3',}
    
    Afterwards, simulate function can be called to simulate the world. This returns a pandas dataframe
    with output values of all FMUs as specified during add_fmu phase.
    
    Adding ability to interact with World with each time step, opeining up the possibility to make the cosimulation object interact 
    with optimization algorithms written in python.
    '''

class World():
    
    def __init__(self, start_time = 0, stop_time = 100, logging = False, exchange = 0, clean_up = True, interpolate_results = True):
        self.start_time = start_time
        self.stop_time = stop_time
        self.fmu_dict = {}
        self.stepsize_dict = {}
        self.powerflow_stepsize_dict = {}
        self.logging = logging
        self.exchange = exchange
        self.modify_signal = False
        self.modify_dict = {}
        self.signal_dict = {}
        self.signals_dict_new = {}
        self.init_dict = {}
        self.clean_up = clean_up
        self.all_outputs = []
        self.powerflow_dict = {}
        self.powerflow_outputs = {}
        self.powerflow_inputs = {}
        self.snapshots_dict = {}
        self.pf = ''
        self.interpolate_results = interpolate_results
        self.G = True
        self.csv_dict = {}
        self.variable_dict = {}

    def add_powerflow(self, network_name, net_loc, inputs = [], outputs = [], pf = 'pf', step_size=900, logger = 'DEBUG'):
        self.stepsize_dict[network_name] = step_size
        assert network_name is not None, "No name specified for power flow model. Name must be specified."
        assert net_loc is not None, "No location specified for power flow model. Can't read without."
        try:
            network = pypsa_adapter(network_name, net_loc, inputs = inputs, outputs = outputs, logger_level = logger)
        except:
            network = pp_adapter(network_name, net_loc, inputs = inputs, outputs = outputs)
        self.powerflow_dict[network_name] = network
        self.pf = pf
        if self.logging:
            print("Added powerflow network '%s' to the world" %(network_name))

    
    def add_fmu(self, fmu_name, fmu_loc, step_size, inputs = [], outputs=[], exist=False, solver_name = 'Cvode', variable=False, **kwargs):
        if 'validate' in kwargs.keys():
            validate=kwargs['validate']
        self.stepsize_dict[fmu_name] = step_size
        m_desc = read_model_description(fmu_loc, validate=validate)
        fmi_type = 'CoSimulation' if m_desc.coSimulation is not None else 'ModelExchange'
        if fmi_type == 'CoSimulation':
            fmu_temp = FmuCsAdapter(fmu_loc,
                             instanceName=fmu_name,
                             step_size = step_size,
                             inputs = inputs,
                             outputs=outputs,
                             exist=exist)
            self.fmu_dict[fmu_name] = fmu_temp
        elif fmi_type == 'ModelExchange':
            fmu_temp = FmuMeAdapter(fmu_loc,
                             instanceName=fmu_name,
                             step_size = step_size,
                             inputs = inputs,
                             outputs=outputs,
                             solver_name = solver_name)
            self.fmu_dict[fmu_name] = fmu_temp
        
        self.all_outputs = [fmu_name + '.' + output for output in outputs]
        self.variable_dict[fmu_name] = variable
        if self.logging:
            print("Added a FMU '%s' to the world" %(fmu_name))
        
        
    
    def add_signal(self, signal_name, signal):
        signal_obj = signal_adapter(signal_name, signal)
        self.signal_dict[signal_name] = signal_obj
        if self.logging:
            print("Added a signal '%s' to the world" %(signal_name))
    
    def add_csv(self, csv_name, csv_location):
        df = pd.read_csv(csv_location)
        #analyse the df, and calculate step size
        if 'time' not in df.columns:
            print('No time column in csv file. Please convert csv file to required format. CSV not added.')
        else:
            autocorr = df.time.autocorr()
            if round(autocorr,4) != 1:
                print('energysim can only read csv with fixed time intervals. Current file does not have time stamps with fixed interval. Cant add csv.')
            else:
                dt = df.time[1] - df.time[0]
                self.stepsize_dict[csv_name] = dt
                self.csv_dict[csv_name] = df
                if self.logging:
                    print(f"Added csv: {csv_name} to World")
                
        
        
    def add_connections(self,connections={}):
        assert (len(self.fmu_dict) + len(self.powerflow_dict) > 0),"Cannot add connections when no FMUs specified!"
        if len(self.fmu_dict)+len(self.signal_dict)+len(self.powerflow_dict) == 1 and len(self.csv_dict) == 0:
            self.connections_between_fmus = {}
        else:
            self.connections_between_fmus = connections
        if self.logging:
            print("Added %i connections to the world" %(len(self.connections_between_fmus)))
    
    def get_fmu_name(self, name):
        if type(name).__name__ == 'str':
            return name.split('.')[0]
        if type(name).__name__ == 'tuple':
            return [x.split('.')[0] for x in name]
        
    def plot(self, plot_edge_labels=False):
        self.G=nx.DiGraph()
        for key, value in self.connections_between_fmus.items():
            
            if type(key).__name__ == 'str' and type(value).__name__ == 'str':
                n1 = self.get_fmu_name(key)
                n2 = self.get_fmu_name(value)
                self.G.add_node(n1)
                self.G.add_node(n2)
                self.G.add_edge(n1, n2, name=key + ' to ' + value)
            elif type(key).__name__ == 'str' and type(value).__name__ == 'tuple':
                n1 = self.get_fmu_name(key)
                self.G.add_node(n1)
                for v in value:
                    n2 = self.get_fmu_name(v)
                    self.G.add_node(n2)
                    self.G.add_edge(n1, n2, name=key + ' to ' + v)
            elif type(key).__name__ == 'tuple' and type(value).__name__ == 'str':
                n2 = self.get_fmu_name(value)
                self.G.add_node(n2)
                for k in n1:
                    n1 = self.get_fmu_name(k)
                    self.G.add_node(n1)
                    self.G.add_edge(n1, n2, name=k + ' to ' + value)
            else:
                print('There is a many to many dependance in the graph. Cannot create graph.')
        pos=nx.spring_layout(self.G)
        nx.draw_networkx_nodes(self.G,pos, node_color='r', alpha=0.8, with_labels=True)
        nx.draw_networkx_edges(self.G,pos, alpha=0.8, edge_color='b')
        nx.draw_networkx_labels(self.G,pos)
        if plot_edge_labels:
            edge_labels = nx.get_edge_attributes(self.G,'name')
            nx.draw_networkx_edge_labels(self.G,pos, edge_labels=edge_labels)
        
        plt.show()
                
        
        
    def get_lcm(self):
        max_order = int(max([len(str(i).split('.')[1]) for i in self.stepsize_dict.values()]))
        new_list = 10**(max_order)*np.array(self.stepsize_dict.values())
        
        from functools import reduce    # need this line if you're using Python3.x
    
        def lcm(a, b):
            if a > b:
                greater = a
            else:
                greater = b
        
            while True:
                if greater % a == 0 and greater % b == 0:
                    lcm = greater
                    break
                greater += 1
        
            return lcm
        
        def get_lcm_for(your_list):
            return reduce(lambda x, y: lcm(x, y), your_list)
        
        final_lcm = get_lcm_for(new_list)
        self.big_time_step = final_lcm/10**max_order
        
        return self.big_time_step
    

    def perform_consistency_name_checks(self):
        a1 = self.csv_dict.keys()
        a2 = self.fmu_dict.keys()
        a3 = self.signal_dict.keys()
        a4 = self.powerflow_dict.keys()
        if len(set(a1).intersection(a2, a3, a4)) != 0:
            return False
        else:
            return True
        
    def create_df_for_pf(self, net_name, network):        
        colums_of_pf = ['time']
        for item in network.outputs:
            name = [net_name + '.' + item]
            colums_of_pf.extend(name)
        
        return pd.DataFrame(columns = colums_of_pf)
    
    def create_df_for_simulator(self, name, obj):
        columns_of_df = ['time']
        for item in obj.outputs:
            name = [name + '.' + item]
            columns_of_df.extend(name)
        return pd.DataFrame(columns = columns_of_df)
    
    def init(self):
        '''
        Initialises energysim object
        '''
        if self.logging:
            print("Simulation started..")
            print("Simulation status:\n")
        
        #create simulator list
        self.simulator_list = {}
        self.simulator_list.update(self.fmu_dict )
        self.simulator_list.update(self.powerflow_dict)
        
        
        self.create_results_dataframe()
        
        assert (len(self.fmu_dict) + len(self.powerflow_dict) > 0),"Cant run simulations when no simulators are specified!"
        if len(self.fmu_dict) + len(self.powerflow_dict) > 1:
            assert (len(self.connections_between_fmus) > 0),"Connections between FMUs are not specified!"
        
        self.res_dict = {}
        
        #initialise FMUs and their result dataframes
        for name, _fmu in self.fmu_dict.items():
            fmu_df = self.create_df_for_fmu(name)            
            self.res_dict[name] = fmu_df
            try:
                _fmu.setup()
                #check if initial vakues need to be set before initilisation
                if self.init_dict:
                    self.set_parameters(self.init_dict)
                    
                self.set_csv_signals(0)
                _fmu.init()
                print('Initialised FMU: %s'%(_fmu.instanceName))
            except:
                try:
                    print("Couldn't initialise %s. Trying again.." %(_fmu.instanceName))
                    _fmu.setup()
                    #check if initial vakues need to be set before initilisation
                    if self.init_dict:
                        self.set_parameters(self.init_dict)
                        
                    self.set_csv_signals(0)
                    _fmu.init()
                except:
                    print('Couldnt initialise %s. Simulation stopped.' %(_fmu.instanceName))
                    sys.exit()
        
        #initialise power flow and create its dataframe as well
        for net_name, network in self.powerflow_dict.items():
            net_df = self.create_df_for_pf(net_name, network)
            self.res_dict[net_name] = net_df
            self.set_csv_signals(0)
            network.init()
            print('Initialised powerflow network: %s'%(net_name))
        
        #determine when FMUs exchange information between themselves
        if self.exchange == 0 and len(self.fmu_dict.items()) + len(self.powerflow_dict.items())>1:
            self.final_tStep = self.get_lcm()
        elif self.exchange == 0 and len(self.fmu_dict.items()) + len(self.powerflow_dict.items()) == 1:
            self.final_tStep = list(self.stepsize_dict.values())[0]
        else:
            self.final_tStep = self.exchange
        
        
        if len(self.fmu_dict)+len(self.signal_dict)+len(self.powerflow_dict) > 1:
            self.do_exchange = True
        else:
            self.do_exchange = False
        
    def simulate(self, startTime=False, stopTime=False):
        '''
        Simulates the energysim object from startTime to stopTime with a step. If no input argument is specified, start, stop, and step times are derived from energysim iniitalization.
        '''
        check = self.perform_consistency_name_checks()
        if not check:
            print('Found more than one similar names for added fmu, signal, powerflow, or csv. Please use unique names for each add_xxx() method.')
            print('Exiting simulation.')
            sys.exit()
        
        if not startTime and not stopTime:
            flag = True
            dis = False
            startTime = self.start_time
            stopTime = self.stop_time
            self.init()
        else:
            flag = False
            dis = None
        assert (stopTime-startTime >= self.final_tStep), "difference between start and stop time > exchange value specified in world initialisation"
        total_steps = int((stopTime-startTime)/self.final_tStep)+1
        
        for time in tqdm(np.linspace(startTime, stopTime, total_steps), disable = dis):
            
            if self.do_exchange:
                self.exchange_values(time)
                
            for _fmu in self.fmu_dict.keys():                
                temp_time = time
                local_stop_time = min(temp_time + self.final_tStep, stopTime)
                while temp_time < local_stop_time:
                    self.set_csv_signals(temp_time)
                    temp_res = [temp_time] + self.fmu_dict[_fmu].getOutput()
                    self.res_dict[_fmu].loc[len(self.res_dict[_fmu].index)] = temp_res
                    
                    if _fmu in self.variable_dict.keys():                        
                        stepsize = self.get_step_time(self.stepsize_dict[_fmu], local_stop_time, temp_time, self.final_tStep, time)
                        try:
                            self.fmu_dict[_fmu].step_advanced(min(temp_time, local_stop_time), stepsize)
                            temp_time += stepsize             #self.stepsize_dict[_fmu]
                        except:
                            print(f'caught exception at time {temp_time}, changing step size from {stepsize} to {self.stepsize_dict[_fmu]}')
                            self.fmu_dict[_fmu].step(min(temp_time, local_stop_time), self.stepsize_dict[_fmu])
                            temp_time += self.stepsize_dict[_fmu]
                                                
                    else:
                        self.fmu_dict[_fmu].step(min(temp_time, local_stop_time))
                        temp_time += self.stepsize_dict[_fmu]
                
            for net_name, network in self.powerflow_dict.items():                
                if round(time%self.stepsize_dict[net_name]) == 0:
                    self.set_csv_signals(time)
                    network.step()
                    
                    temp_res = [time] + network.getOutput()
                    self.res_dict[net_name].loc[len(self.res_dict[net_name].index)] = temp_res
        
        if flag:
            self.clean_canvas()
            return self.results()
            
    def clean_canvas(self):            
        if self.clean_up:
            try:
                for _fmu in self.fmu_dict.values():
                    _fmu.cleanUp()
            except:
                print('Tried deleting temporary FMU files, but failed. Try manually.')
            for _net in self.powerflow_dict.values():
                _net.cleanUp()
        
    def sync(self, time):
        '''
        Manual call for exchange of values between simulators.
        '''
        self.exchange_values(time)
        
    def set_value(self, parameters, value):
        '''
        Must specify parameters and values in list format
        '''
        for item, val in zip(parameters, value):
            sim_name = item.split('.')[0]
            variable = item.replace(sim_name,'')[1:]
            self.simulator_list[sim_name].set_value([variable], [val])


    def get_value(self, parameter):
        '''
        Must specify parameter in a list format.
        '''
        outputs = []
        for item in parameter:
            sim_name = item.split('.')[0]
            variable = item.replace(sim_name,'')[1:]
            temp_value = self.simulator_list[sim_name].get_value([variable])
            outputs.append(temp_value[0])
        return outputs
    
    def get_new_time(self, stop, curr):        
        return (stop-curr)/2
        
    def get_step_time(self, step_size, local_stop_time, current_time, final_tStep, int_start_time):
        if current_time-int_start_time < 0.01 or local_stop_time-current_time < 0.01:
            new_step_size = 0.001
        else:
            new_step_size = final_tStep-0.01
        return new_step_size
    
    def results(self):
        if self.interpolate_results:
            try:
                self.process_results(self.res_dict)
                return self.results_dataframe
            except:
                return self.res_dict
        else:
            return self.res_dict
     
    
    def process_results(self, results_dict):
        if self.interpolate_results:
            from scipy.interpolate import interp1d
            temp1 = [(x, len(y)) for x, y in results_dict.items()]
            temp2 = dict(temp1)
            most_fmu_tsteps = max(temp2, key=temp2.get)
            self.new_tp = np.array(results_dict[most_fmu_tsteps].loc[:, 'time'])
            self.new_time_points = np.linspace(self.start_time, self.stop_time, len(np.asarray(self.new_tp <= self.stop_time).nonzero()[0]))
            
            self.results_dataframe.loc[:,'time'] = self.new_time_points
            
            #go through the results dict and create interpolated values.
            #define interpolation funciton
            for key, dataframe in results_dict.items():
                df_time = dataframe.loc[:,'time']
                df_time_temp = np.linspace(self.start_time, self.stop_time, len(np.asarray(df_time <= self.stop_time).nonzero()[0]))
                del df_time
                rem_columns = list(dataframe.columns)
                rem_columns.remove('time')
                for column_name in rem_columns:
                    y = np.array(dataframe.loc[:,column_name])[:len(df_time_temp)]
                    f = interp1d(df_time_temp, y, kind='previous')
                    new_datavalues = f(self.new_time_points)
                    self.results_dataframe.loc[:,column_name] = new_datavalues
        else:
            pass

        
    def create_df_for_fmu(self, fmu_name):
        colums_of_fmu = ['time']
        _fmu = self.fmu_dict[fmu_name]
        for item in _fmu.outputs:
            name = [_fmu.instanceName + '.' + item]
            colums_of_fmu.extend(name)
        
        fmu_df = pd.DataFrame(columns = colums_of_fmu)
        
        return fmu_df
    
        
    def set_csv_signals(self, t):
        for output_, input_ in self.connections_between_fmus.items():
            is_csv = True if type(output_).__name__ == 'str' and output_.split('.')[0] in self.csv_dict.keys() else False
            if is_csv:           # and type(output_).__name__ == 'str':
                
                if type(input_).__name__ == 'str':
                    input_ele_name = input_.split('.')[0]
                    is_powerflow = True if input_ele_name in self.powerflow_dict.keys() else False
                    is_fmu = True if input_ele_name in self.fmu_dict.keys() else False
                    if is_fmu:
                        input_fmu = self.fmu_dict[input_ele_name]
                        input_variable = input_.replace(input_ele_name,'')[1:]
                        
                        csv_name = output_.split('.')[0]
                        csv_variable = output_.split('.')[1]
                        csv_dt = self.stepsize_dict[csv_name]
                        temp_var = self.csv_dict[csv_name].at[int(t/csv_dt), csv_variable]
                        input_fmu.set_value([input_variable], [temp_var])
        
                    if is_powerflow:
                        network = self.powerflow_dict[input_ele_name]
                        network_variable = input_.replace(input_ele_name,'')[1:]
                        
                        csv_name = output_.split('.')[0]
                        csv_variable = output_.split('.')[1]
                        csv_dt = self.stepsize_dict[csv_name]
                        temp_var = self.csv_dict[csv_name].at[int(t/csv_dt), csv_variable]
                        network.set_value([network_variable], [temp_var])
                    
                if type(input_).__name__ == 'tuple':
                    for input__ in input_:
                        input_ele_name = input__.split('.')[0]
                        is_powerflow = True if input_ele_name in self.powerflow_dict.keys() else False
                        is_fmu = True if input_ele_name in self.fmu_dict.keys() else False
                        
                        if is_fmu:
                            input_fmu = self.fmu_dict[input_ele_name]
                            input_variable = input__.replace(input_ele_name,'')[1:]
                            
                            csv_name = output_.split('.')[0]
                            csv_variable = output_.split('.')[1]
                            csv_dt = self.stepsize_dict[csv_name]
                            temp_var = self.csv_dict[csv_name].at[int(t/csv_dt), csv_variable]
                            input_fmu.set_value([input_variable], [temp_var])
            
            
                        if is_powerflow:
                            network = self.powerflow_dict[input_ele_name]
                            network_variable = input__.replace(input_ele_name,'')[1:]
                            
                            csv_name = output_.split('.')[0]
                            csv_variable = output_.split('.')[1]
                            csv_dt = self.stepsize_dict[csv_name]
                            temp_var = self.csv_dict[csv_name].at[int(t/csv_dt), csv_variable]
                            
                            network.set_value([network_variable], [temp_var])
    
    def get_output_exchange(self, op_, t):
        ele_name = op_.split('.')[0]
        is_fmu = True if ele_name in self.fmu_dict.keys() else False
        is_pp_net = True if ele_name in self.powerflow_dict.keys() else False
        is_signal = True if ele_name in self.signal_dict.keys() else False
        is_csv = True if ele_name in self.csv_dict.keys() else False
        
        if is_fmu:
            output_fmu = self.fmu_dict[ele_name]
            output_variable =op_.replace(ele_name,'')[1:]
            temp_var = output_fmu.get_value([output_variable])[0]
        if is_pp_net:
            network = self.powerflow_dict[ele_name]
            output_variable =op_.replace(ele_name,'')[1:]
            temp_var = network.get_value([output_variable])[0]
        if is_signal:
            signal = self.signal_dict[ele_name]
            temp_var = signal.get_value(t)
        if is_csv:
            self.set_csv_signals(t)
            temp_var = 'its csv'
            
        if self.modify_dict:
            flag = True if op_ in self.modify_dict.keys() else False
            if flag:
                if len(self.modify_dict[op_]) == 1:
                    ret_output = temp_var*self.modify_dict[op_][0]
                elif len(self.modify_dict[op_]) == 2:
                    ret_output = temp_var*self.modify_dict[op_][0] + self.modify_dict[op_][1]
                else:
                    print('Unknown signal modification on output %s. Not applying modification.' %(op_))
                    ret_output = temp_var
            else:
                ret_output = temp_var
        else:
            ret_output = temp_var
        
        return ret_output
        
    
        
    def exchange_values(self, t):        
        for output_, input_ in self.connections_between_fmus.items():
            if type(output_).__name__ == 'str':
                temp_var = self.get_output_exchange(output_, t)
            elif type(output_).__name__ == 'tuple':
                temp_list = []
                for item in output_:
                    t_v = self.get_output_exchange(item, t)
                    temp_list.append(t_v)
                temp_var = sum(temp_list)
            
            if not type(temp_var).__name__ == 'str':
                if type(input_).__name__ == 'str':
                    ele_name = input_.split('.')[0]
                    pp_net = True if ele_name in self.powerflow_dict.keys() else False
                    fmu = True if ele_name in self.fmu_dict.keys() else False
                    
                    if fmu:
                        input_fmu = self.fmu_dict[ele_name]
                        input_variable = input_.replace(ele_name,'')[1:]
                        input_fmu.set_value([input_variable], [temp_var])
                    if pp_net:
                        input_pp = self.powerflow_dict[ele_name]
                        input_variable = input_.replace(ele_name,'')[1:]
                        input_pp.set_value([input_variable], [temp_var])
                    
                elif type(input_).__name__ == 'tuple':
                    for item in input_:
                        ele_name = item.split('.')[0]
                        pp_net = True if ele_name in self.powerflow_dict.keys() else False
                        fmu = True if ele_name in self.fmu_dict.keys() else False
                        
                        if fmu:
                            input_fmu = self.fmu_dict[ele_name]
                            input_variable = item.replace(ele_name,'')[1:]
                            input_fmu.set_value([input_variable], [temp_var])
                        if pp_net:
                            input_pp = self.powerflow_dict[ele_name]
                            input_variable = item.replace(ele_name,'')[1:]
                            input_pp.set_value([input_variable], [temp_var])
                    
    def create_results_dataframe(self):
        self.list_of_columns = ['time']
        for _fmu in self.fmu_dict.values():
            for item in _fmu.outputs:
                name = [_fmu.instanceName + '.' + item]
                self.list_of_columns.extend(name)
        self.results_dataframe = pd.DataFrame(columns = self.list_of_columns)
    
    
    def options(self, settings):
        assert (len(self.fmu_dict) + len(self.powerflow_dict) > 0),"Cannot add settings to world when no FMUs are specified!"
        
        if 'parameter_set' in settings.keys():
            self.set_parameters(settings['parameter_set'])
        
        if 'modify_signal' in settings.keys():
            self.modify_dict = settings['modify_signal']
        
        if 'init' in settings.keys():
            self.init_dict = settings['init']
    
    
    def set_parameters(self, parameter_dictionary):
        _sims = parameter_dictionary.keys()
        for _sim in _sims:
            parameter_list_derived = list(parameter_dictionary[_sim])[0]
            parameter_values_to_set = list(parameter_dictionary[_sim])[1]
            if _sim in self.fmu_dict.keys():
                self.fmu_dict[_sim].set_value(parameter_list_derived, parameter_values_to_set)
            if _sim in self.powerflow_dict.keys():
                self.powerflow_dict[_sim].set_value(parameter_list_derived, parameter_values_to_set)
            
        
    def plot_results(self, xlab = 'Time(s)', ylab = '', y = [], scientific = False):
        locol = [x for x in self.list_of_columns if x!='time']
        plt.figure(2)
        if self.interpolate_results:
            time_vector = self.results_dataframe.time
            if len(y) == 0:
                for item in locol:
                    plt.plot(time_vector,self.results_dataframe.loc[:,item], label=item)
            else:
                for item in y:
                    try:
                        plt.plot(time_vector,self.results_dataframe.loc[:,item], label=item)
                    except:
                        print("Could not find the variable %s in results dataframe. Make sure you have \
                              specified the variables as 'fmu.variable'. Skipping." %(item))
                        continue
            plt.ticklabel_format(useOffset=scientific)      
            plt.xlabel(xlab)
            plt.ylabel(ylab)
            plt.legend()
            plt.show()
        else:
            print('Results can only be plotted natively if interpolate_results option is set to True in World() options.')


        
        


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  3 16:13:38 2018

@author: digvijaygusain
"""
import os, sys
from fmpy import read_model_description, extract, dump
from fmpy.fmi2 import *
from fmpy.util import plot_result, download_test_file, auto_interval
from fmpy import supported_platforms
import shutil
from fmpy.simulation import Recorder, Input, apply_start_values
from random import random


class FmuCsAdapter():
    '''
    FMU CoSimulation adapter for energysim
    '''
    
    eps = 1.0e-13
    
    def __init__(self, fmu_location, 
                 instanceName=None, 
                 start_time=0, 
                 tolerance=1e-06, 
                 stop_time = 100, 
                 step_size = 1.0e-3, 
                 inputs = [], 
                 outputs = [],
                 show_fmu_info = False,
                 exist=False):
        assert (fmu_location is not None), "Must specify FMU location"
        self.fmu_location = fmu_location
        if instanceName is None:
            instanceID = int(random()*1000)
            self.instanceName = 'fmu'+str(instanceID)
            print(self.instanceName)
        else:
            self.instanceName = instanceName
        self.exist = exist
        self.tolerance = tolerance
        self.start_time = start_time
        self.stop_time = stop_time
        self.output_interval = step_size
        self.inputs = inputs
        self.outputs = outputs
        if show_fmu_info:
            dump(self.fmu_location)
        self.setup()

        
    def setup(self):
        self.t_next = self.start_time
        if self.exist:
            self.unzipDir = self.fmu_location
        else:
            self.unzipDir = extract(self.fmu_location)
        self.modelDescription = read_model_description(self.fmu_location, validate=True)
        self.is_fmi1 = self.modelDescription.fmiVersion == '1.0'
        
        logger = printLogMessage
        
        if self.is_fmi1:
            callbacks = fmi1CallbackFunctions()
            callbacks.logger = fmi1CallbackLoggerTYPE(logger)
            callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
            callbacks.freeMemory = fmi1CallbackFreeMemoryTYPE(freeMemory)
            callbacks.stepFinished = None
        else:
            callbacks = fmi2CallbackFunctions()
            callbacks.logger = fmi2CallbackLoggerTYPE(logger)
            callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
            callbacks.freeMemory = fmi2CallbackFreeMemoryTYPE(freeMemory)
        
        #define var values for input and output variables
        self.vrs = {}
        for variable in self.modelDescription.modelVariables:
            self.vrs[variable.name] = [variable.valueReference, variable.type]

            
        if self.is_fmi1:
            self.fmu = FMU1Slave(guid = self.modelDescription.guid,
                                 unzipDirectory=self.unzipDir,
                                 modelIdentifier=self.modelDescription.coSimulation.modelIdentifier,
                                 instanceName=self.instanceName)
            self.fmu.instantiate(functions=callbacks)
        else:
            self.fmu = FMU2Slave(guid = self.modelDescription.guid,
                     unzipDirectory=self.unzipDir,
                     modelIdentifier=self.modelDescription.coSimulation.modelIdentifier,
                     instanceName=self.instanceName)
            self.fmu.instantiate(callbacks=callbacks)
        
        self.input = Input(self.fmu, self.modelDescription, None)
        
        

    def set_value(self,parameterName,Value):
        '''
        Must specify parameters and values in list format
        '''
        for i, j in zip(parameterName, Value):
            
            if self.vrs[i][1] == 'Real':
                self.fmu.setReal([self.vrs[i][0]], [j])
            elif self.vrs[i][1] in ['Integer', 'Enumeration']:
                self.fmu.setInteger([self.vrs[i][0]], [j])
            elif self.vrs[i][1] == 'Boolean':
                if isinstance(j, str):
                    if j.lower() not in ['true', 'false']:
                        raise Exception('The value "%s" for variable "%s" could not be converted to Boolean' %
                                        (j, i))
                    else:
                        j = j.lower() == 'true'
                self.fmu.setBoolean([self.vrs[i][0]], [bool(j)])
            elif self.vrs[i][1] == 'String':
                self.fmu.setString([self.vrs[i][0]], [j])
    

    def get_value(self,parameterName):
        '''
        Must specify parameter in a list format.
        '''
        values_ = []
        for i in parameterName:
            if self.vrs[i][1] == 'Real':
                temp = self.fmu.getReal([self.vrs[i][0]])
            elif self.vrs[i][1] in ['Integer', 'Enumeration']:
                temp = self.fmu.getInteger([self.vrs[i][0]])
            elif self.vrs[i][1] == 'Boolean':
                temp = self.fmu.getBoolean([self.vrs[i][0]])
            elif self.vrs[i][1] == 'String':
                temp = self.fmu.getString([self.vrs[i][0]])
            
            values_.append(temp[0])
        return values_
        
    def reset(self):
        self.fmu.reset()
        
    def init(self): 
        if self.is_fmi1:
            self.fmu.initialize()
        else:
            self.fmu.setupExperiment(startTime=self.start_time, tolerance=self.tolerance)
            self.fmu.enterInitializationMode()
            self.fmu.exitInitializationMode()
    
    
    def set_inital_inputs(self, starting_values):
        from fmpy.simulation import apply_start_values
        apply_start_values(fmu = self.fmu,
                           model_description = self.modelDescription,
                           start_values = starting_values,
                           apply_default_start_values=False)
        
    def setInput(self, inputValues):
        self.inputVariables = []
        for i in self.inputs:
            self.inputVariables.append(self.vrs[i][0])
        self.fmu.setReal(list(self.inputVariables), list(inputValues))
    
    def getOutput(self):
        self.outputVariables = []
        for i in self.outputs:
            self.outputVariables.append(self.vrs[i][0])
        
        return self.fmu.getReal(list(self.outputVariables))
            
    
    def step_advanced(self, time, step_size=None):        
        #step ahead in time
        self.input.apply(time)
        if step_size is None:
            self.fmu.doStep(currentCommunicationPoint = time, communicationStepSize = self.output_interval)
        else:
            self.fmu.doStep(currentCommunicationPoint = time, communicationStepSize = step_size)
# TODO -           while a!=0:
#                self.fmu.setFMUstate(state)
##                print(f"Didnt work with stepsize = {step_size}, new stepsize = {step_size/2}.")
#                step_size = step_size/2
#                a = self.fmu.doStep(currentCommunicationPoint = time, communicationStepSize = step_size)
#            self.fmu.freeFMUstate(state)                    
#            print(f'returning to master: status = {a}, step size = {step_size}.')
#        return a, step_size
    
    def step_v2(self,time, stepsize):
        self.input.apply(time)
        return self.fmu.doStep(currentCommunicationPoint = time, communicationStepSize = stepsize)
    
    def step(self, time):
        #step ahead in time
        self.input.apply(time)
        return self.fmu.doStep(currentCommunicationPoint = time, communicationStepSize = self.output_interval)
    
    def cleanUp(self):
        self.fmu.terminate()
        self.fmu.freeInstance()
        
        shutil.rmtree(self.unzipDir)
    
    def simulate(self, timeout=180):
        from fmpy import simulate_fmu
            
        result = simulate_fmu(self.fmu_location, 
                              start_time = self.start_time, 
                              stop_time =self.stop_time, 
                              timeout = timeout, 
                              output_interval = self.output_interval, 
                              output = self.outputs)
        return result
        



# -*- coding: utf-8 -*-
"""
Created on Sat Jul  6 15:07:17 2019

@author: Digvijay

Pandapower adapter for FMUWorld
"""


import pandapower as pp, sys

class pp_adapter():
    
    def __init__(self, network_name, net_loc, inputs = [], outputs = []):
        '''
        Initialises the pandapower network adapter for energysim cosimulaiton object. 
        Specify the following:
                network_name: unique name for your network
                net_loc: location of the pandapower network file (specified as pickle file)
                inputs: list of all inputs. Specified as object_name.variable
                outputs: list of all output variables to record.
        '''
        self.network_name = network_name
        self.net_loc = net_loc
        self.network = pp.from_pickle(net_loc)
        all_names_exist = self.check_names(self.network)
        assert all_names_exist, "All load, generator, sgen, bus, and external grid elements must have names. Simulation stopped"
        print('Successfully imported network')
        self.new_inputs, self.new_outputs = self.process_powerflow_ipop(self.network, inputs, outputs)
        self.outputs = outputs

    def check_names(self, network):
        for n in [ 'load', 'sgen', 'ext_grid', 'bus']:
            check = getattr(getattr(network, n),"name").any()
            if not check:
                return False
        return True
    
    def init(self):
        pp.runpp(self.network)
    
    def set_value(self, parameters, values):
        '''
        Must specify parameters and values in list format
        '''
        for parameter, value in zip(parameters, values):
            ele_name, input_variable = parameter.split('.')
            assert input_variable in ['P', 'Q'], "Powerflow input variable not valid. Use P, Q to  define variables."
            if ele_name in list(self.network.gen.name):
                adder, residual = 'gen', 'p_mw' if input_variable == 'P' else 'q_mvar'
            elif ele_name in list(self.network.load.name):
                adder, residual = 'load', 'p_mw' if input_variable == 'P' else 'q_mvar'
            elif ele_name in list(self.network.sgen.name):
                adder, residual = 'sgen', 'p_mw' if input_variable == 'P' else 'q_mvar'
            else:
                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
                sys.exit()
            
            getattr(self.network, adder).at[(getattr(self.network, adder).name == ele_name).idxmax(), residual] = value
    
    def get_value(self, parameters):
        '''
        Must specify parameter in a list format.
        '''
        temp_parameter_list = [x.split('.') for x in parameters]
        temp_output = []
        for ele_name, output_variable in temp_parameter_list:
            if output_variable.lower() == 'v':
                x = 'vm_pu'
            elif output_variable.lower() == 'va':
                x = 'va_degree'
            if ele_name in list(self.network.gen.name):
                adder, residual = 'res_gen', 'p_mw' if output_variable == 'P' else 'q_mvar'
            elif ele_name in list(self.network.load.name):
                adder, residual = 'res_load', 'p_mw' if output_variable == 'P' else 'q_mvar'
            elif ele_name in list(self.network.bus.name):
                adder, residual = 'res_bus', x
            elif ele_name in list(self.network.sgen.name):
                adder, residual = 'res_sgen', 'p_mw' if output_variable == 'P' else 'q_mvar'
            else:
                print(f'Could not find {ele_name} as a component in {self.network} network. Make sure element names are correctly specified in get_value()')
                sys.exit()
                
            temp_var = getattr(self.network, adder).at[(getattr(self.network, adder[4:]).name == ele_name).idxmax(), residual]
            temp_output.append(temp_var)
        
        return temp_output
        
    
    def getOutput(self):
        return [getattr(self.network, adder).at[(getattr(self.network, adder[4:]).name == ele).idxmax(), residual] for ele, adder, residual in self.new_outputs]
        
    
    def setInput(self, inputValues):
        pass
    
    def step(self):  
        a = pp.runpp(self.network)
        return a
    
    def process_powerflow_ipop(self, network, inputs, outputs):
            new_inputs = []
            new_outputs = []
            for item in inputs:
                ele_name, input_variable = item.split('.')
                assert input_variable in ['P', 'Q'], "Powerflow input variable not valid. Use P, Q to  define variables."
                #check in generators
                if ele_name in list(network.gen.name):
                    adder, residual = 'gen', 'p_mw' if input_variable == 'P' else 'q_mvar'
                elif ele_name in list(network.load.name):
                    adder, residual = 'load', 'p_mw' if input_variable == 'P' else 'q_mvar'
                elif ele_name in list(network.sgen.name):
                    adder, residual = 'sgen', 'p_mw' if input_variable == 'P' else 'q_mvar'
                else:
                    print(f'Only Generator, load, and sgen P, Q inputs are supported for pandapower nets. Couldnt find {ele_name} in either loads or generators. Quitting simulation.')
                    sys.exit()
                new_inputs.append((ele_name, adder, residual))
            
            for item in outputs:
                ele_name, output_variable = item.split('.')
                assert output_variable in ['P', 'Q', 'V', 'Va'], "Powerflow output variable not valid. Use P, Q, V, Va to  define variables."
                
                if output_variable == 'V':
                    x = 'vm_pu'
                elif output_variable == 'Va':
                    x = 'va_degree'
                    
                if ele_name in list(network.gen.name):
                    adder, residual = 'res_gen', 'p_mw' if output_variable == 'P' else 'q_mvar'
                elif ele_name in list(network.load.name):
                    adder, residual = 'res_load', 'p_mw' if output_variable == 'P' else 'q_mvar'
                elif ele_name in list(network.bus.name):
                    adder, residual = 'res_bus', x
                elif ele_name in list(network.sgen.name):
                    adder, residual = 'res_sgen', 'p_mw' if output_variable == 'P' else 'q_mvar'
                else:
                    print(f'Only Generator, load, storage, and bus P, Q, outputs are supported. Couldnt find {ele_name} in specified network. Quitting simulation.')
                    sys.exit()
                new_outputs.append((ele_name, adder, residual))
                
            return new_inputs, new_outputs   
    
    def cleanUp(self):
        del self.network


from setuptools import setup

def readme():
    with open('README.md', encoding='utf-8') as f:
        README = f.read()
    return README


setup(
      name='energysim',
      version='0.0.2',
      description='Python package for cosimulation of FMU based multi-energy systems',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://github.com/dgusain1/energysim',
      author='Digvijay Gusain',
      author_email='d.gusain@tudelft.nl',
      license='MIT',
      classifiers=[
              'License :: OSI Approved :: MIT License',
              'Programming Language :: Python :: 3',
              'Programming Language :: Python :: 3.6'],
      packages=['energysim'],
      include_package_data=True,
      install_requires=['numpy', 'pandas', 'tqdm', 'fmpy', 'pandapower', 'pypsa']     
      )


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  6 16:44:38 2019

@author: digvijaygusain
signal_adapter
"""

class signal_adapter():
    def __init__(self, signal_name, signal):
        self.signal_name = signal_name
        self.signal = signal
        
    
    def get_value(self, time):
        
        if len(self.signal) == 1:
            return self.signal[0]
        elif len(self.signal) == 3:
            return self.signal[0] if time < self.signal[1] else self.signal[2]
        elif len(self.signal) == 4:
            return self.signal[0] if time < self.signal[1] or time > self.signal[2] else self.signal[3]
        else:
            print('unrecognized signal format, using 0 values for signal %s.' %(self.signal_name))
            return 0